# pylint: disable=unused-argument
""" A graphical user interface for WFEngine based on ipywidgets """
import ast
import os
import os.path
import time
from pathlib import Path
import ipywidgets as widgets
from ipywidgets import HBox, VBox, Layout
from ipyfilechooser import FileChooser
from IPython.display import display, clear_output, JSON
from fireworks import LaunchPad
from fireworks.user_objects.queue_adapters.common_adapter import CommonAdapter
from fireworks.fw_config import LAUNCHPAD_LOC, QUEUEADAPTER_LOC
from fireworks import Workflow
from wfengine import WFEngine
from wfengine_remote import WFEngineRemote
from BUILTIN import transfer_files


# Getting the arguments with FileChooser and SelectMultiple
lp_file = FileChooser('')
qa_file = FileChooser('')
wf_file = FileChooser('')

style = {'description_width': 'initial'}

# Set default paths for file chooser
HOME = str(Path.home())
lp_file.show_hidden = True
lp_file.default_path = os.path.join(HOME, '.fireworks')
lp_file.default_filename = 'launchpad.yaml'
qa_file.show_hidden = True
qa_file.default_path = os.path.join(HOME, '.fireworks')
qa_file.default_filename = 'qadapter.yaml'
wf_file.default_path = os.path.join(HOME)
wf_file.default_filename = 'workflow.yaml'

# Show folder icons for file chooser
lp_file.use_dir_icons = True
qa_file.use_dir_icons = True

# Set multiple file filter patterns for config file
lp_file.filter_pattern = ['*.yaml', '*.json']
qa_file.filter_pattern = ['*.yaml', '*.json']
wf_file.filter_pattern = ['*.yaml', '*.json']

# Set titles for choosers
lp_file.title = 'Select your launchpad file'
qa_file.title = 'Select your qadapter file'
wf_file.title = 'Select your workflow file'


text_input_layout = Layout(width='100%')

configure_button = widgets.Button(
    tooltip='Create a new and manage running engine',
    description='Manage engine',
    style=style,
    icon='')
output_configure_button = widgets.Output()

rsync_button = widgets.Button(
    tooltip='Create a new and manage running engine',
    description='Transfer files',
    style=style,
    icon='')
output_rsync_button = widgets.Output()

send_files_button = widgets.Button(
    tooltip='Create a new and manage running engine',
    description='Transfer files',
    style=style,
    icon='')

output_send_files_button = widgets.Output()


plus_lib = widgets.Button(tooltip='Transfer additional module(s) to remote machines', icon='plus', style=style, layout=Layout(width='100px'), description='Modules', title= 'add modules')
plus_lib_output = widgets.Output()


plus_input = widgets.Button(tooltip='Transfer additional files(s) to remote machines', icon='plus', style=style, layout=Layout(width='100px'), description='Files', title= 'add files')
plus_input_output = widgets.Output()


clear_lib_button = widgets.Button(tooltip='Clear module list', icon='refresh', style=style, layout=Layout(width='40px'), description='')
clear_lib_button_output = widgets.Output()

clear_input_button = widgets.Button(tooltip='Clear input list', icon='refresh', style=style, layout=Layout(width='40px'), description='')
clear_input_button_output = widgets.Output()



lpad_button = widgets.Button(
    description='Load launchpad',
    style=style,
    icon='upload')
output_lpad_button = widgets.Output()

qadapter_button = widgets.Button(
    description='Load qadapter',
    style=style,
    icon='upload')
output_qadapter_button = widgets.Output()

new_engine_button = widgets.Button(
    tooltip='New engine will be created',
    description='Create new engine',
    style=style,
    icon='cogs')
output_new_engine_button = widgets.Output()

dump_engine_button = widgets.Button(
    tooltip='Save the current engine into a file',
    description='Save engine',
    style=style,
    icon='download')

resume_engine_button = widgets.Button(
    tooltip='Load an engine from a file',
    description='Load engine',
    style=style,
    icon='upload')

output_saveload_engine_button = widgets.Output()

manage_launcher_button = widgets.Button(
    tooltip='Manage launcher threads',
    description='Manage launcher',
    style=style,
    icon='')
output_manage_launcher_button = widgets.Output()

start_button = widgets.Button(
    description='Start launcher',
    style=style,
    icon='play')
output_start_button = widgets.Output()

stop_button = widgets.Button(
    description='Stop launcher',
    style=style,
    icon='power-off')
output_stop_button = widgets.Output()

manage_workflows_button = widgets.Button(
    tooltip='Manage workflows',
    description='Manage workflows',
    style=style,
    icon='')
output_manage_workflows_button = widgets.Output()

new_workflow_button = widgets.Button(
    tooltip='Add a new workflow to the engine',
    description='Add new workflow',
    style=style,
    icon='plus')
output_new_workflow_button = widgets.Output()

add_workflow_button = widgets.Button(
    description='Add workflow',
    style=style,
    icon='upload')
output_add_workflow_button = widgets.Output()

remove_workflow_button = widgets.Button(
    description='Remove a workflow',
    style=style,
    icon='trash')
output_remove_workflow_button = widgets.Output()

manage_nodes_button = widgets.Button(
    tooltip='Manage individual workflow nodes',
    description='Manage nodes',
    style=style,
    icon='')
output_manage_nodes_button = widgets.Output()

status_button = widgets.Button(
    tooltip='Show the status of all workflows',
    description='Status',
    style=style,
    icon='eye')
output_status_button = widgets.Output()

status_button_detail = widgets.Button(
    tooltip='Show the status of specified nodes',
    description='Status',
    style=style,
    icon='search-plus')
status_button_detail_output = widgets.Output()

rerun_node_button = widgets.Button(
    description='Rerun a node',
    style=style,
    icon='redo')
output_rerun_node_button = widgets.Output()

query = widgets.Textarea(
    value='None',
    placeholder='{}',
    description='Query',
    tooltip='Query to select the workflows to include to engine',
    layout=text_input_layout,
    disabled=False)

engine_file = FileChooser('')
engine_file.default_path = '.'
engine_file.use_dir_icons = True
engine_file.filter_pattern = ['*.yaml', '*.json']
engine_file.title = 'WFEngine file'

firework_id = widgets.IntText(
    value=1,
    description='fw id',
    placeholder='firework id',
    disabled=False
)

configure_engine_method = widgets.RadioButtons(
    options=['Save or load an existing engine',
             'Create a new engine from scratch'],
    value='Save or load an existing engine',
    layout={'width': 'max-content'},
    description='Method',
    disabled=False
)
configure_engine_method_output = widgets.Output()

remote_cluster = widgets.Checkbox(
    value=False,
    description='Use a remote cluster',
    tooltip='Use a remote cluster to launch jobs',
    disabled=False,
    indent=False
)
remote_cluster_output = widgets.Output()

user_name = widgets.Textarea(
    value='',
    placeholder='xy1234',
    description='Username',
    tooltip='Username on the remote cluster',
    layout=text_input_layout,
    disabled=False)

host_name = widgets.Textarea(
    value='',
    placeholder='hostname.domainname',
    description='Hostname',
    tooltip='Hostname of the remote cluster',
    layout=text_input_layout,
    disabled=False)

remote_conf = widgets.Textarea(
    value='',
    placeholder='module load xyz',
    description='Command',
    tooltip='Remote configuration command',
    layout=text_input_layout,
    disabled=False)

launch_dir = widgets.Textarea(
    value='.',
    placeholder='/absolute/path/to/launchdir',
    description='Launchdir',
    tooltip='Directory in that nodes will be executed',
    layout=text_input_layout,
    disabled=False)

dest_dir = widgets.Textarea(
    value='.',
    placeholder='/absolute/path/to/destdir',
    description='Destination of transfer, default is launchdir if it is set',
    tooltip='Directory in that nodes will be executed',
    layout=text_input_layout,
    disabled=False)

eng_name = widgets.Textarea(
    value=None,
    placeholder='None',
    description='Engine name',
    tooltip='A name for the engine (one will be generated if not specified)',
    layout=text_input_layout,
    disabled=False)



libpath = []
def plus_lib_clicked(b3):
    libpath.append(FileChooser(''))
    libpath[len(libpath)-1].default_path = '../'
    libpath[len(libpath)-1].use_dir_icons = True
    with plus_lib_output:
        display(libpath[len(libpath)-1])
        
inputpath = []
def plus_input_clicked(b3):
    inputpath.append(FileChooser(''))
    inputpath[len(inputpath)-1].default_path = '../'
    inputpath[len(inputpath)-1].use_dir_icons = True
    with plus_input_output:
        display(inputpath[len(inputpath)-1])
        
        

def clear_lib_button_clicked(b):
    with plus_lib_output:
        clear_output()
        libpath.clear() 
        
def clear_input_button_clicked(b):
    with plus_input_output:
        clear_output()
        inputpath.clear() 
    with output_send_files_button:
        clear_output()
        
        
         
def clear_engineoutput():
    """ Clear outputs """
    with output_manage_nodes_button:
        clear_output()
    with output_new_workflow_button:
        clear_output()
    with output_new_workflow_button:
        clear_output()
    with output_stop_button:
        clear_output()
    with output_start_button:
        clear_output()
    with output_saveload_engine_button:
        clear_output()


def clear_consoleoutput():
    """ clear outputs """
    with output_status_button:
        clear_output()
    with output_remove_workflow_button:
        clear_output()
    with output_rerun_node_button:
        clear_output()


def clear_button_outputs():
    """ Clear top buttons output """
    with output_new_workflow_button:
        clear_output()
    with output_configure_button:
        clear_output()
    with output_rsync_button:
        clear_output()
    with output_manage_launcher_button:
        clear_output()
    with output_manage_workflows_button:
        clear_output()
    with output_stop_button:
        clear_output()
    with output_manage_nodes_button:
        clear_output()
    with output_new_engine_button:
        clear_output()
    with output_saveload_engine_button:
        clear_output()
    clear_consoleoutput()


def manage_nodes_button_clicked(bvar):
    """ Manage nodes button is clicked """
    with output_manage_nodes_button:
        clear_button_outputs()
        clear_output()
        display(firework_id)
        display(HBox([status_button, status_button_detail,
                      rerun_node_button]), output_status_button,
                status_button_detail_output, output_rerun_node_button)


def configure_button_clicked(bvar):
    """ Configure button is clicked """
    with output_configure_button:
        clear_button_outputs()
        clear_output()
        display(VBox([configure_engine_method,
                      configure_engine_method_output]))
        configure_engine_method_changed(bvar)


def manage_launcher_button_clicked(bvar):
    """ manage launcher button is clicked """
    with output_manage_workflows_button:
        clear_button_outputs()
        clear_output()
        display(HBox([start_button, stop_button]), output_start_button,
                output_stop_button)

        
def rsync_button_clicked(bvar):
    """ Prepare to send files"""
    with output_rsync_button:
        clear_button_outputs()
        clear_output()
        display(user_name,host_name, dest_dir)
        display(HBox([plus_input, clear_input_button]), plus_input_output)
        display(send_files_button, output_send_files_button)
        
def send_files_button_clicked(bvar):
    """ Transfer files clicked"""
    with output_send_files_button:
        clear_output()
        transfer_files(user_name.value, host_name.value, inputpath, dest_dir.value)
        
    
        
def manage_workflows_button_clicked(bvar):
    """ manage workflows button is clicked """
    with output_manage_workflows_button:
        clear_button_outputs()
        clear_output()
        display(firework_id)
        display(HBox([status_button, new_workflow_button,
                      remove_workflow_button]), output_status_button,
                output_new_workflow_button, output_remove_workflow_button)


def new_workflow_button_clicked(bvar):
    """ new workflow button is clicked """
    with output_new_workflow_button:
        clear_output()
        display(wf_file)
        display(add_workflow_button, output_add_workflow_button)


def remote_cluster_changed(bvar):
    """ toggle the remote cluster checkbox """
    with remote_cluster_output:
        clear_output()
        if remote_cluster.value and configure_engine_method.value == 'Create a new engine from scratch':
            display(user_name, host_name, remote_conf)
            display(HBox([plus_lib, clear_lib_button]),
                    plus_lib_output)

            


def configure_engine_method_changed(bvar):
    """ select engine configuration method from radio buttons """
    with configure_engine_method_output:
        clear_output()
        if configure_engine_method.value == 'Save or load an existing engine':
            print('Save or load an existing engine')
            display(VBox([remote_cluster, remote_cluster_output]))
            display(engine_file)
            display(HBox([dump_engine_button, resume_engine_button]),
                    output_saveload_engine_button)
            if remote_cluster.value:
                remote_cluster_changed(bvar)
        if configure_engine_method.value == 'Create a new engine from scratch':
            print('Create a new engine from scratch')
            display(lp_file)
            display(lpad_button, output_lpad_button)
            display(qa_file)
            display(qadapter_button, output_qadapter_button)
            display(query, launch_dir, eng_name)
            display(VBox([remote_cluster, remote_cluster_output]))
            remote_cluster_changed(bvar)
            display(new_engine_button, output_new_engine_button)


class WFEnginejupyter():
    """ A class for construcing a GUI for FireWorks """
    wfe = None

    def __init__(self):
        """ Load default launchpad and qadapter """
        if LAUNCHPAD_LOC:
            try:
                self.jlaunchpad = LaunchPad.from_file(LAUNCHPAD_LOC)
                print('Default lpad in ' + LAUNCHPAD_LOC + ' is loaded')
            except FileNotFoundError as error:
                print('The default launchpad file is missing.')
                print('Error message is:' + str(error.args))
            except KeyError as error:
                print('The default lpad file is not correct.')
                print('Error message is:' + str(error.args))

        if QUEUEADAPTER_LOC:
            try:
                self.jqadapter = CommonAdapter.from_file(QUEUEADAPTER_LOC)
                print('Default qadapter in ' + QUEUEADAPTER_LOC + ' is loaded')
            except FileNotFoundError as error:
                print('The default qadapter file is missing.')
                print(error.args)
            except KeyError as error:
                print('The default qadapter file is not correct.')
                print(error.args)
    
    def new_engine_button_clicked(self, bvar):
        """ create new engine button is clicked """
        with output_new_engine_button:
            clear_output()
            if self.wfe and self.wfe.thread and self.wfe.thread.is_alive():
                self.wfe.stop()
                print('Stopping the running launcher thread, please wait ...')
                self.wfe.thread.join()
            else:
                query_dict = ast.literal_eval(query.value)
                if remote_cluster.value:
                    assert host_name.value != '' and user_name.value != ''
                    self.wfe = WFEngineRemote(launchpad=self.jlaunchpad,
                                              qadapter=self.jqadapter,
                                              wf_query=query_dict,
                                              host=host_name.value,
                                              user=user_name.value,
                                              conf=remote_conf.value,
                                              libpath = libpath,
                                              launchdir=launch_dir.value,
                                              name=eng_name.value,
                                             )
                    print('Engine is created for remote launcher')
                else:
                    self.wfe = WFEngine(launchpad=self.jlaunchpad,
                                        qadapter=self.jqadapter,
                                        wf_query=query_dict,
                                        launchdir=launch_dir.value,
                                        name=eng_name.value)
                    print('Engine is created for local launcher')

    def resume_engine_button_clicked(self, bvar):
        """ resume engine button is clicked """
        with output_saveload_engine_button:
            clear_output()
            if self.wfe and self.wfe.thread and self.wfe.thread.is_alive():
                self.wfe.stop()
                print('Stopping the running launcher thread, please wait ...')
                self.wfe.thread.join()
            eng_class = WFEngineRemote if remote_cluster.value else WFEngine
            try:
                self.wfe = eng_class.from_file(filename=engine_file.value)
            except (FileNotFoundError, IsADirectoryError) as error:
                print('Specify a valid engine filename or path.')
                print(error.args)
            except ValueError as error:
                if 'Unsupported format' in error.args[0]:
                    print('Specify a valid engine filename ending with .json or .yaml')
                    print(error.args)
                else:
                    raise
            except PermissionError as error:
                print(error.args)
            else:
                print('An engine has been loaded from file', engine_file.value)

    def dump_engine_button_clicked(self, bvar):
        """ dump the engine to file """
        with output_saveload_engine_button:
            clear_output()
            if self.wfe:
                try:
                    self.wfe.to_file(filename=engine_file.value)
                except (FileNotFoundError, IsADirectoryError) as error:
                    print('Specify a valid engine filename or path.')
                    print(error.args)
                except ValueError as error:
                    if 'Unsupported format' in error.args[0]:
                        print('Specify a valid engine filename ending with .json or .yaml')
                        print(error.args)
                    else:
                        raise
                except PermissionError as error:
                    print(error.args)
                else:
                    print('The engine has been saved in file', engine_file.value)
            else:
                print('Create an engine first.')

    def lpad_button_clicked(self, bvar):
        """ load user defined launchpad """
        with output_lpad_button:
            clear_output()
            if lp_file.selected is None:
                print('Select a launchpad file.')
                time.sleep(1)
                return
            try:
                self.jlaunchpad = LaunchPad.from_file(lp_file.selected)
            except FileNotFoundError as error:
                print('Select a valid launchpad file')
                print(error.args)
            else:
                print('Launchpad has been loaded from file', lp_file.selected)

    def qadapter_button_clicked(self, bvar):
        """ load user defined qadapter """
        with output_qadapter_button:
            clear_output()
            if qa_file.selected is None:
                print('Select a qadapter file.')
                time.sleep(1)
                return
            try:
                self.jqadapter = CommonAdapter.from_file(qa_file.selected)
            except FileNotFoundError as error:
                print('Select a valid qadapter file')
                print(error.args)
            else:
                print('Qadapter has been loaded from file', qa_file.selected)

    def add_wf_clicked(self, bvar):
        """ add a workflow from the file """
        with output_add_workflow_button:
            clear_output()
            if wf_file.selected is None:
                print('Select a workflow file.')
                time.sleep(1)
                return
            try:
                workflow = Workflow.from_file(wf_file.selected)
                self.wfe.add_workflow(workflow=workflow)
            except FileNotFoundError as error:
                print('Select a valid workflow file')
                print(error.args)
            else:
                print('Workflow has been loaded from file', wf_file.selected)

    def start_launcher_clicked(self, bvar):
        """ start launcher button clicked """
        clear_engineoutput()
        with output_start_button:
            clear_output()
            self.wfe.start()
            print('The launcher thread is created.')


    def stop_launcher_clicked(self, bvar):
        """ stop launcher button clicked """
        clear_engineoutput()
        with output_stop_button:
            try:
                self.wfe.stop()
                print('please wait ...')
                self.wfe.thread.join()
                clear_output()
            except AttributeError:
                clear_output()
                print('There is no running thread')

    def status_button_clicked(self, bvar):
        """ status summary """
        clear_consoleoutput()
        with output_status_button:
            clear_output()
            self.wfe.status_summary()

    def status_detailed_button_clicked(self, bvar):
        """ status details about specified nodes """
        clear_consoleoutput()
        with output_status_button:
            status_dict = self.wfe.status_detail(firework_id.value)
            del status_dict['_id']
            for element in status_dict['archived_launches']:
                del element['_id']
            status_dict['previous_launches']=status_dict['archived_launches']    
            del status_dict['archived_launches']
            del status_dict['launches'][0]['_id']
            display(JSON(status_dict))
            

    def remove_workflow_button_clicked(self, bvar):
        """ remove a workflow from engine """
        clear_consoleoutput()
        with output_remove_workflow_button:
            self.wfe.remove_workflow(firework_id.value)
            self.wfe.status_summary()

    def rerun_node_button_clicked(self, bvar):
        """ rerun a node and print a new status summary """
        clear_consoleoutput()
        with output_rerun_node_button:
            self.wfe.rerun_node(firework_id.value)
            self.wfe.status_summary()


wfengine = WFEnginejupyter()

new_engine_button.on_click(wfengine.new_engine_button_clicked)
manage_workflows_button.on_click(manage_workflows_button_clicked)
rsync_button.on_click(rsync_button_clicked)
send_files_button.on_click(send_files_button_clicked)
resume_engine_button.on_click(wfengine.resume_engine_button_clicked)
configure_button.on_click(configure_button_clicked)
new_workflow_button.on_click(new_workflow_button_clicked)
lpad_button.on_click(wfengine.lpad_button_clicked)
qadapter_button.on_click(wfengine.qadapter_button_clicked)
start_button.on_click(wfengine.start_launcher_clicked)
stop_button.on_click(wfengine.stop_launcher_clicked)
status_button.on_click(wfengine.status_button_clicked)
status_button_detail.on_click(wfengine.status_detailed_button_clicked)
dump_engine_button.on_click(wfengine.dump_engine_button_clicked)
manage_launcher_button.on_click(manage_launcher_button_clicked)
manage_workflows_button.on_click(manage_workflows_button_clicked)
add_workflow_button.on_click(wfengine.add_wf_clicked)
remove_workflow_button.on_click(wfengine.remove_workflow_button_clicked)
manage_nodes_button.on_click(manage_nodes_button_clicked)
rerun_node_button.on_click(wfengine.rerun_node_button_clicked)
remote_cluster.observe(remote_cluster_changed)
configure_engine_method.observe(configure_engine_method_changed)
plus_lib.on_click(plus_lib_clicked)
plus_input.on_click(plus_input_clicked)
clear_lib_button.on_click(clear_lib_button_clicked)
clear_input_button.on_click(clear_input_button_clicked)



def display_wfengine():
    display(HBox([configure_button, manage_launcher_button,
                  manage_workflows_button, manage_nodes_button, rsync_button]),
            output_configure_button, output_manage_launcher_button,
            output_manage_workflows_button, output_manage_nodes_button, output_rsync_button)



import numpy as np
import time
import os
import subprocess

def callscript(*args, **kwargs):
    cmd = kwargs['command']
    if 'argument' in kwargs.keys():
        cmd = cmd + " " + str(kwargs['argument'])
    os.system(cmd)
    return 'Done!!! '


def printrun(*args):
    result = []
    for arg in args:
        if isinstance(arg, list) and len(arg) == 1:
            result.append(arg[0])
        else:
            result.append(arg)
    if len(result) == 1:
        result = result[0]
    print(result)
    return result


def MERGE(**kwargs):
    dic_merged = {}
    for key, value in kwargs.items():
        if isinstance(value, np.number):
            value = float(value)
        else:
            value = value
    return (kwargs)   


def flat_tuple(lazy_tuple, return_number):
    flat_list = []
    for i in range(return_number):
        flat_list.append([])
    for i in range(return_number):       
        for j in range(len(lazy_tuple)):    
            for item in lazy_tuple[j][i]:  
                flat_list[return_number- (i + 1)].append(item)
    return flat_list


def flat_list(lazy_list):
    flat_list = []
    for list in lazy_list:
            for item in list:
                flat_list.append(item)
    return flat_list


def wfgenes_argparser(args):

    """ Check if the arguments are provided from the command-line. 
    If there is no command-line value for any argument, it will be loaded from WConfig"""
    resource = {}        
    resource['pool'] = args.pool 
    if resource['pool'] == 'local_threads':
        resource['threads_number'] = args.threads_number
    elif resource['pool'] == 'slurm':
        resource['cpu_per_node']  = args.cpu_per_node     
        resource['scale'] = args.scale     
        resource['worker_per_node'] = args.worker_per_node     
        resource['memory'] = args.memory     
        resource['partition'] = args.partition      
        resource['walltime'] = args.walltime    
    elif resource['pool'] == 'local_processes':
        resource['processes_number'] = args.processes_number 
        resource['cpu_per_process'] = args.cpu_per_process
    
    return resource

        
def transfer_files(user, host, filespath, launchdir, subfolder='', path_type='PYTHONPATH'):
    """ Transfer a list of file in the given path """
    space = ' '
    ssh_link = user + '@' + host + ':'
    subfolder_path = os.path.join(launchdir, subfolder)
    env_path = 'export ' + path_type + '=$' + path_type + ':' + subfolder_path
    rsync_cmd = "rsync -a --rsync-path='mkdir -p"
    rsync_cmd += space + subfolder_path + space
    rsync_cmd += "&& rsync'" + space
    if len(filespath) != 0:
        for path in filespath:
            if path.selected_path is not None:
                rsync_extend = ''
                if not path.selected_filename:
                    rsync_extend = rsync_cmd
                    rsync_extend += "-rv" + space
                    rsync_extend += path.selected_path + '/*' + space
                    rsync_extend += ssh_link
                    rsync_extend += subfolder_path
                    print(rsync_extend)
                    proc = subprocess.call(rsync_extend, shell=True)
                else:
                    rsync_extend = rsync_cmd
                    rsync_extend += "-v" + space
                    rsync_extend += path.selected + space
                    rsync_extend += ssh_link
                    rsync_extend += subfolder_path
                    print(rsync_extend)
                    proc = subprocess.call(rsync_extend, shell=True)
    return env_path


        



""" wfGenes: Automatic workflow generator."""


__author__ = 'Mehdi Roozmeh'
__email__ = 'mehdi.roozmeh@kit.edu'
__copyright__ = 'Copyright 2020, Karlsruhe Institute of Technology'


import os
import os.path
import sys
from copy import deepcopy
from collections import OrderedDict
import argparse
import json
import xmlschema
import yaml
import fireworks_schema
from wfgenes.core.initial_wfgenes import BasewfGenes
from wfgenes.core.pywrapper_wfgenes import PywrapperwfGens
from wfgenes.core.simstack_wfgenes import SimstackwfGenes
from wfgenes.core.fireworks_wfgenes import FireworkwfGenes
from wfgenes.core.dot_wfgenes import DotwfGenes
from wfgenes.core.clean_wfgenes import CleanwfGenes
from wfgenes.core.dask_wfgenes import DaskwfGenes
from wfgenes.core.parsl_wfgenes import ParslwfGenes
import time
import signal
import jsonschema
from jsonschema import validate
import subprocess



def keyboardInterruptHandler(signal, frame):
    print("KeyboardInterrupt (ID: {}) has been caught. Cleaning up...".format(signal))
    exit(0)

def wconfig_validator(BasewfGenes, args, wconfig ):
    last_error = ''
    verified = 'null'
    i = 'null'
    #time.sleep(1) # Avoid extra cpu load
    try:
        wfgenes_init = BasewfGenes(args)
        verified = True
        i = 'recent_verified'
        return wfgenes_init , verified, i
    except jsonschema.ValidationError as e:
        if e.message != last_error:
            last_error = e.message
            print(e.instance)
            print (last_error)
            verified = False
            print('Press ctrl+c to exit or modify ' + wconfig)
            return 'not_verified' , verified, i


def modified(wconfig, wconfig_stat):
    if wconfig_stat[0] == 'null' and wconfig_stat[1] == 'null':
        wconfig_stat[1] = os.stat(wconfig).st_mtime
        wconfig_stat[0] = wconfig_stat[1]
        return True        
    elif wconfig_stat[0] == wconfig_stat[1]:
        wconfig_stat[1] = os.stat(wconfig).st_mtime
        return False
    else:
        wconfig_stat[0] = wconfig_stat[1]
        return True
    
def glue_args(*blueargs):
    argdic = {"workflowconfig": "", "inputpath":"", "outputpath":"", "wms":""}
    if bool(blueargs) == False:      
        parser = argparse.ArgumentParser()
        parser.add_argument(
            '--workflowconfig',
            default='workflow.yaml',
            help='workflowconfig file contains data of each'
            ' routine in yaml format. i.e.'
            ' Input/Outputfile, modules and arguments list, the default is workflow.yaml.'
            ' For more information check README file.')

        parser.add_argument('--inputpath', default='',
                            help='Set input directory of WMS.'
                            ' It is not necessary for all WMS,'
                            ' please refer to related manual in repository.')


        parser.add_argument('--wms', default='all',
                            help='Choose specific workflow management system.'
                            'The default behavior produces all supported WMSs.')

        parser.add_argument('--embedded', default='False',
                    help='If True the generated script for Dask/Parsl should be run by FireWorks')
                    
        args = parser.parse_args()

        argdic ["workflowconfig"] = args.workflowconfig
        argdic ["inputpath"] = args.inputpath
        argdic ["wms"] = args.wms
        argdic ["embedded"] = args.embedded


    elif bool(blueargs) == True:
        pass
    return argdic

def run_wfgenes(args):
    png = ''
    wconfig = args['workflowconfig']
    wconfig_stat = ['null', 'null']
    signal.signal(signal.SIGINT, keyboardInterruptHandler)   
    i = 'null' # Entrance token to wconfig_validator func
    valid_wms = {'FireWorks','SimStack', 'Dask','Parsl'}
    if args['wms'] in valid_wms:
        print("wfGenes start converting WConfig to", args['wms'])
    else:    
        print("The selected argument for --wms is not in the list. Posiible options.", valid_wms)
        exit()    
    while True:
        # while user using wfGenes
        time.sleep(1) # Avoid extra cpu load
        if modified(wconfig, wconfig_stat) == True:
            #init= BasewfGenes (args)
            validator = wconfig_validator(BasewfGenes, args ,wconfig)
            wfgenes_init = validator[0]
            verified =  validator[1]
            i =  validator[2]
            if verified == True:
                model = wfgenes_init.wconfig_checker()
                wconfig_status = model[0]
                wconfig_version = model[2]
                if wconfig_status == 'active': 
                    # modeling phase
                    print('Model is verified and Graph generation is active. Remove "active" suffix to proceed with code generation or continue with modling')
                    i = 'null'
                    print('Press ctrl+c to exit')
                    try:
                        graphics = DotwfGenes(wfgenes_init)
                        png = graphics.png_file
                    except Exception as error:
                        print('unexpected error:'+error.args[0])

                else:
                    try:
                        graphics = DotwfGenes(wfgenes_init)
                        png = graphics.png_file
                    except Exception as error:
                        print('unexpected error:'+error.args[0])
                    try : 
                        if args['wms'] == 'FireWorks' or args['wms'] == 'all':
                            graphics = DotwfGenes(wfgenes_init)
                            png = graphics.png_file    
                            fireworks_dict=FireworkwfGenes(wfgenes_init)
                            print('wfGenes succeeded with model generation using FireWorks')
                    except FileNotFoundError as error:
                        print('wfGenes failed with model generation using FireWorks')
                        print(error)
                        pass
                    except Exception as error:
                        print('wfGenes failed with model generation using FireWorks')
                        print('unexpected error:'+ error.args[0])   
                    try:            
                        if args['wms'] == 'SimStack' or args['wms'] == 'all':
                            graphics = DotwfGenes(wfgenes_init)
                            png = graphics.png_file                            
                            simstack_wrapper = PywrapperwfGens(wfgenes_init, args)
                            test2=SimstackwfGenes(args)
                            print('wfGenes succeeded with model generation using SimStack')

                    except Exception as error:
                        print('wfGenes failed with model generation using SimStack')
                        print('unexpected error:'+error.args[0])       
                    try:
                        if args['wms'] == 'Dask' or args['wms'] == 'all':  
                            graphics = DotwfGenes(wfgenes_init)
                            png = graphics.png_file    
                            dask_wf = DaskwfGenes(wfgenes_init, args) 
                            print('wfGenes succeeded with model generation using Dask')
                    except Exception as error:
                        print('wfGenes failed with model generation using Dask')
                        print('unexpected error:'+ error.args[0])
                    try:
                        if args['wms'] == 'Parsl' or args['wms'] == 'all':
                            graphics = DotwfGenes(wfgenes_init)
                            png = graphics.png_file    
                            parsl_wf= ParslwfGenes(wfgenes_init, args)
                            print('wfGenes succeeded with model generation using Parsl')
                    except Exception as error:
                        print('wfGenes failed with model generation using Parsl')
                        print('unexpected error:' + error.args[0])

                    print("Outputs is saved in:", wfgenes_init.wconfig_checker()[1] )
                    return png


            
def main():
    start_time= time.time()
    arg_dic = glue_args()
    png= run_wfgenes(arg_dic)
    end_time = time.time()
    execution_time = end_time - start_time 
    print("wfGenes finished ",arg_dic["wms"], "in",execution_time,"seconds")

def gui():
    cmd = 'jupyter-lab ../wfgenes_gui.ipynb'
    proc = subprocess.call(cmd, shell=True)
    

if __name__ == '__main__':
    main()
                
        
    
   


    


"""
Methods to manipulate Z-matrices. Example for a Z-matrix:
[
    {
        'label': 'c1',
        'symbol': 'c'
    },
    {
        'label': 'h1',
        'symbol': 'h',
        'distance': {'reference': 'c1', 'value': 1.3, 'constant': False}
    },
    {
        'label': 'h2',
        'symbol': 'h',
        'distance': {'reference': 'c1', 'value': 1.3},
        'angle': {'reference': 'h1', 'value': 120.0}
    },
    {
        'label': 'h3',
        'symbol': 'h',
        'distance': {'reference': 'c1', 'value': 1.3},
        'angle': {'reference': 'h2', 'value': 120.0},
        'dihedral': {'reference': 'h1', 'value': 180.0, 'constant': True}
    }
]
"""
import numpy as np
from numpy import pi, sin, cos, arccos, sqrt
from ase import Atoms
from ase.data import chemical_symbols

@classmethod
def from_zmat(cls, *args, **kwargs):
    dct = zmat2cart(*args, **kwargs)
    dct['constraint'] = zmat2constr(*args)
    return cls(**dct)

Atoms.from_zmat = from_zmat

def zmat2cart(zmat, positions=None):
    """
    Convert a Z-matrix to Cartesian coordinates
    The Cartesian coordinates of the first three atoms can be supplied in the
    optinal 'positions' parameter. Otherwise the first atom will be at the
    origin, the second atom displaced in z direction, and the third atom
    constrained in the yz-plane. For all remaining atoms the algorithm from
    C. R. Eddy, J. Chem. Phys. 38, 1032 (1963) is used.
    """

    labels = {atom['label']: index for index, atom in enumerate(zmat)}
    symbols = [atom['symbol'].capitalize() for atom in zmat]
    assert len(labels) == len(zmat), 'repeated label'
    assert all(s in chemical_symbols for s in symbols), 'unknown symbol'

    if positions is None:
        positions = [[0., 0., 0.]]
        if len(zmat) == 1:
            return {'symbols': symbols, 'positions': positions}
        stre1 = zmat[1]['distance']['value']
        positions.append([0., 0., stre1])
        if len(zmat) == 2:
            return {'symbols': symbols, 'positions': positions}
        stre2 = zmat[2]['distance']['value']
        bend2 = zmat[2]['angle']['value']*pi/180.0
        if labels[zmat[2]['distance']['reference']] == 1:
            assert labels[zmat[2]['angle']['reference']] == 0
            posy = stre2*sin(bend2)
            posz = stre1-stre2*cos(bend2)
        else:
            assert labels[zmat[2]['distance']['reference']] == 0
            assert labels[zmat[2]['angle']['reference']] == 1
            posz = stre2*cos(bend2)
            posy = sqrt(stre2**2-posz**2)
        positions.append([0., posy, posz])

    for zline in zmat[3:]:
        A = labels[zline['dihedral']['reference']]
        B = labels[zline['angle']['reference']]
        C = labels[zline['distance']['reference']]
        alpha = np.array(positions[B])-np.array(positions[A])
        beta = np.array(positions[C])-np.array(positions[B])
        alpha = alpha/np.linalg.norm(alpha)
        beta = beta/np.linalg.norm(beta)
        lambd = [(alpha[(j+1)%3]*beta[(j+2)%3]-beta[(j+1)%3]*alpha[(j+2)%3])
                 /sin(arccos(np.dot(alpha, beta))) for j in range(3)]
        multiplier1 = cos(pi-zline['dihedral']['value']*pi/180.0)
        multiplier2 = sin(pi-zline['dihedral']['value']*pi/180.0)
        mu = [lambd[j]*multiplier1
              +(lambd[(j+1)%3]*beta[(j+2)%3]-beta[(j+1)%3]*lambd[(j+2)%3])
              *multiplier2 for j in range(3)]
        multiplier1 = cos(pi-zline['angle']['value']*pi/180.0)
        multiplier2 = sin(pi-zline['angle']['value']*pi/180.0)
        gamma = [beta[j]*multiplier1
                 +(beta[(j+1)%3]*mu[(j+2)%3]-mu[(j+1)%3]*beta[(j+2)%3])
                 *multiplier2 for j in range(3)]
        pos = [positions[C][j]+zline['distance']['value']*gamma[j] for j in range(3)]
        positions.append(pos)

    return {'symbols': symbols, 'positions': positions}


def zmat2constr(zmat):
    """ Construct ASE constraint object from a Z-matrix """
    from ase.constraints import FixInternals
    labels = {zl['label']: zln for zln, zl in enumerate(zmat)}
    assert len(labels) == len(zmat), 'repeated label'
    bconstr = []
    aconstr = []
    dconstr = []
    for zln, zl in enumerate(zmat):
        if 'distance' in zl:
            disref = labels[zl['distance']['reference']]
            disval = zl['distance']['value']
            if zl['distance'].get('constant', False):
                bind = [disref, zln]
                bconstr.append([disval, bind])
        if 'angle' in zl:
            angref = labels[zl['angle']['reference']]
            angval = zl['angle']['value']
            if zl['angle'].get('constant', False):
                aind = [angref, disref, zln]
                aconstr.append([angval*pi/180., aind])
        if 'dihedral' in zl:
            dihref = labels[zl['dihedral']['reference']]
            dihval = zl['dihedral']['value']
            if zl['dihedral'].get('constant', False):
                dind = [dihref, angref, disref, zln]
                dconstr.append([dihval*pi/180., dind])

    if len(bconstr)+len(aconstr)+len(dconstr):
        return FixInternals(bonds=bconstr, angles=aconstr, dihedrals=dconstr)



""" Functions for modeling charge transport in organic semiconductors for
    sequential workflow execution
"""
__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2016, Karlsruhe Institute of Technology'

from util import atoms2dict
from util import dict2atoms

def prepare_input(parameters):
    """
    input: parameters dict
    output: simulation dict
    """
    from ase.io import read

    title = 'prepare input'
    print('Processing step:', title)

    properties = {}
    simulation = {}
    simulation['morphology'] = {}
    simulation['parameters'] = parameters
    simulation['properties'] = properties

    if 'morphology file' in parameters:
        morphology = read(str(parameters['morphology file']), index=':')
        molecules = []
        for molecule in morphology:
            molecules.append(atoms2dict(molecule, plain_arrays=True))
        simulation['morphology']['molecules'] = molecules
        simulation['monomer structure'] = molecules[0]
    elif 'dimer file' in parameters:
        morphology = read(str(parameters['dimer file']), index=':')
        molecules = []
        for molecule in morphology:
            molecules.append(atoms2dict(molecule, plain_arrays=True))
        simulation['morphology']['molecules'] = molecules
        simulation['optimized monomer structure'] = molecules[0]
    elif 'monomer file' in parameters:
        monomer = read(str(parameters['monomer file']))
        simulation['monomer structure'] = atoms2dict(monomer, plain_arrays=True)
    else:
        raise RuntimeError('morphology / monomer file not found')
    if parameters['carrier'] == 'electron':
        parameters['ion charge'] = parameters['charge']-1
    if parameters['carrier'] == 'hole':
        parameters['ion charge'] = parameters['charge']+1

    return simulation


def optimize_geometry(simulation):
    """
    input: simulation dictionary
    output: simulation dictionary
    """
    from calculators import calculate

    parameters = simulation['parameters']
    monomer = dict2atoms(simulation['monomer structure'], plain_arrays=True)

    title = 'geometry optimization'
    print('Processing step:', title)

    loc_parameters = parameters['calculator']['parameters'].copy()
    loc_parameters['task'] = 'optimize'
    calculate(monomer, parameters['calculator']['name'], loc_parameters)
    simulation['optimized monomer structure'] = atoms2dict(monomer, plain_arrays=True)

    return simulation


def generate_morphology(simulation):
    """
    input: simulation dictionary
    output: simulation dictionary
    """
    from ase import Atoms
    from ase.io import write
    from ase.units import _amu
    from molecules import Morphology
    from morphology import deposit

    monomer = dict2atoms(simulation['monomer structure'], plain_arrays=True)
    parameters = simulation['parameters']

    title = 'generate morphology'
    print('Processing step:', title)

    morphology_type = parameters['morphology type']
    morphology_scale = parameters['morphology scale']
    if morphology_scale == 'atomistic':
        monomer = dict2atoms(
            simulation['optimized monomer structure'],
            plain_arrays=True
        )
    if morphology_type == 'amorphous' and morphology_scale == 'atomistic':
        molecules = deposit(monomer, parameters)
        morphology = Morphology.from_molecules(molecules)
    elif morphology_type == 'crystalline' and morphology_scale == 'coarse-grained':
        if parameters['crystal lattice'] == 'simple cubic':
            morphology = Morphology.SimpleCubic(
                parameters['lattice parameter'],
                parameters['supercell size']
            )
    else:
        raise RuntimeError('unknown morphology type')

    # morphology analysis
    volume = morphology.get_volume()
    mass = morphology.get_mass()
    if mass is not None:
        density_gccm = (mass/volume)*(_amu*1000*1.e24) # 1 cm^3 = 1.e24 Ang^3, _amu in kg
        simulation['properties']['density'] = {'units': 'g/cm^3', 'value': density_gccm}
    [rdf, dist] = morphology.get_rdf(
        parameters['rdf cutoff radius'],
        parameters['rdf number of bins']
    )
    simulation['properties']['rdf'] = [dist.tolist(), rdf.tolist()]

    # plot the RDF
    import matplotlib
    matplotlib.use('Agg') # Force matplotlib to not use any Xwindows backend
    import matplotlib.pyplot as plt
    plt.plot(dist, rdf)
    plt.title(parameters['title'])
    plt.xlabel('r, Ang')
    plt.ylabel('g(r)')
    plt.savefig('morphology-rdf.png')

    # visualize morphology
    if morphology_scale == 'atomistic':
        structure = Atoms()
        for molecule in molecules:
            structure += molecule
        write('morphology-structure.xyz', structure)
        write('morphology-structure.png', structure)
        write('morphology-structure.html', structure)
        write('morphology.xyz', molecules)

    # output
    simulation['morphology'] = {}
    if morphology_scale == 'atomistic':
        m_array = []
        for molecule in molecules:
            m_array.append(atoms2dict(molecule, plain_arrays=True))
        simulation['morphology']['molecules'] = m_array
    else:
        simulation['morphology'] = morphology.todict()
    simulation['properties']['number of molecules'] = len(morphology)

    return simulation


def find_pairs(simulation):
    """
    input: simulation dictionary
    output: simulation dictionary
    """
    from transport import get_pairs
    from molecules import Morphology

    parameters = simulation['parameters']
    if 'molecules' in simulation['morphology']:
        molecules = []
        for molecule in simulation['morphology']['molecules']:
            molecules.append(dict2atoms(molecule, plain_arrays=True))
        morphology = Morphology.from_molecules(molecules)
    else:
        morphology = Morphology(**simulation['morphology'])

    title = 'find pairs'
    print('Processing step:', title)
    assert len(morphology) > 1, 'too few or no molecules in morphology'

    [hopping_sites, hopping_pairs] = get_pairs(morphology, parameters['hopping threshold'])
    simulation['hopping sites'] = hopping_sites
    simulation['hopping pairs'] = hopping_pairs

    simulation['properties']['number of hopping sites'] = len(simulation['hopping sites'])
    simulation['properties']['number of hopping pairs'] = len(simulation['hopping pairs'])

    assert len(hopping_sites) > 0, 'no hopping sites found'
    return simulation


def reorganization_energy(simulation):
    """
    input: simulation dictionary
    output: simulation dictionary
    """
    from transport import reorg_energy

    parameters = simulation['parameters']
    molecule = dict2atoms(simulation['optimized monomer structure'], plain_arrays=True)

    title = 'reorganization energy'
    print('Processing step:', title)

    if 'reorganization energy' in parameters:
        lambda_t = parameters['reorganization energy']
        print('reorganization energy read from input')
    else:
        lambda_t = reorg_energy(structure=molecule, parameters=parameters)
    simulation['properties']['reorganization energy'] = {
        'value': lambda_t,
        'units': 'eV'
    }
    return simulation


def site_energies(simulation):
    """
    input: simulation dictionary
    output: simulation dictionary
    """
    import numpy as np
    from calculators import get_energy

    molecules = []
    for molecule in simulation['morphology']['molecules']:
        molecules.append(dict2atoms(molecule, plain_arrays=True))
    parameters = simulation['parameters']
    calc_params = parameters['calculator']['parameters']
    calc_name = parameters['calculator']['name']
    hopping_sites = simulation['hopping sites']

    if len(hopping_sites) < 1:
        raise ValueError('list hopping_sites has zero length')

    title = 'site energies'
    print('Processing step:', title)

    for site in hopping_sites:
        site_atoms = molecules[site['index']]
        site['energy'] = get_energy(site_atoms, calc_name, calc_params)
        if 'electric field' in parameters:
            q = parameters['ion charge']-parameters['charge']
            site['energy'] += -q*np.dot(
                parameters['electric field'],
                site['center of mass']
            )
        calc = site_atoms.get_calculator()
        eigenvectors = calc.get_mos()[0]
        [number, energy] = calc.get_mo('HOMO')
        site['HOMO'] = {
            'energy': energy,
            'number': number,
            'eigenvector': eigenvectors[number-1].tolist()
        }
        [number, energy] = calc.get_mo('LUMO')
        site['LUMO'] = {
            'energy': energy,
            'number': number,
            'eigenvector': eigenvectors[number-1].tolist()
        }

    return simulation


def electronic_couplings(simulation):
    """
    input: simulation dictionary
    output: simulation dictionary
    """
    from transport import get_e_coupling

    morphology = []
    for molecule in simulation['morphology']['molecules']:
        morphology.append(dict2atoms(molecule, plain_arrays=True))
    parameters = simulation['parameters']

    assert len(simulation['hopping pairs']) > 0, 'hopping_pairs has zero length'

    title = 'electronic couplings'
    print('Processing step:', title)
    for pair in simulation['hopping pairs']:
        pair['coupling'] = get_e_coupling(
            left_index=pair['left'],
            right_index=pair['right'],
            parameters=parameters,
            morphology=morphology,
            hopping_sites=simulation['hopping sites'])

    return simulation


def analyses(simulation):
    """
    input: simulation dictionary
    output: simulation dictionary
    """
    from ase import Atoms
    from ase.io import write
    from transport import marcus_rates, mobility

    morphology = []
    for molecule in simulation['morphology']['molecules']:
        morphology.append(dict2atoms(molecule, plain_arrays=True))

    title = 'charge transfer rates'
    print('Processing step:', title)
    marcus_rates(simulation)

    title = 'carrier mobility'
    print('Processing step:', title)
    mobility(simulation)

    # visualization
    import matplotlib
    matplotlib.use('Agg') # Force matplotlib to not use any Xwindows backend
    import matplotlib.pyplot as plt
    if 'rdf' in simulation['properties']:
        [dist, rdf] = simulation['properties']['rdf']
        # plot the RDF
        plt.plot(dist, rdf)
        plt.title(simulation['parameters']['title'])
        plt.xlabel('r, Ang')
        plt.ylabel('g(r)')
        plt.savefig('morphology-rdf.png')

    # visualize morphology
    structure = Atoms()
    for molecule in morphology:
        structure += molecule
    write('morphology-structure.xyz', structure)
    write('morphology-structure.png', structure)
    write('morphology-structure.html', structure)
    write('morphology.xyz', morphology)

    return simulation


def graph_analysis(simulation):
    """
    input: simulation dictionary
    output: simulation dictionary
    """
    from molecules import Morphology
    from transport import connected_component_analysis
    from transport import MorphologyGraph

    title = 'graph analysis'
    print('Processing step:', title)

    if 'molecules' in simulation['morphology']:
        molecules = []
        for molecule in simulation['morphology']['molecules']:
            molecules.append(dict2atoms(molecule, plain_arrays=True))
        morphology = Morphology.from_molecules(molecules)
    else:
        morphology = Morphology(
            positions=simulation['morphology']['positions'],
            pbc=simulation['morphology']['pbc'],
            cell=simulation['morphology']['cell']
        )

    cutoff = simulation['parameters']['maxflow cutoff']
    direction = simulation['parameters']['maxflow direction']
    method = simulation['parameters']['maxflow method']
    weights = simulation['parameters']['weights mode']

    results = {}

    if weights == 'couplings':
        results.update(
            connected_component_analysis(
                simulation['morphology'],
                simulation['hopping pairs']
            )
        )

    gr = MorphologyGraph(
        morphology,
        weights=weights,
        hopping_pairs=simulation['hopping pairs'],
        parameters=simulation['parameters']
    )

    ka = gr.get_kirchhoff_index(mode='admittance', normalized=True)
    results['kirchhoff admittance'] = ka

    kr = gr.get_kirchhoff_index(mode='resistance', normalized=True)
    results['kirchhoff resistance'] = kr

    flow = gr.get_total_maxflow(cutoff, direction=direction, method=method, color='blue')
    results['maxflow value'] = flow.value
    results['maxflow cut'] = flow.cut
    results['maxflow flow'] = flow.flow
    results['maxflow partition'] = flow.partition

    simulation['properties'].update(results)

    return simulation




""" wfGenes: Automatic workflow generator."""


__author__ = 'Mehdi Roozmeh'
__email__ = 'mehdi.roozmeh@kit.edu'
__copyright__ = 'Copyright 2020, Karlsruhe Institute of Technology'




import os
import os.path
from copy import deepcopy
from collections import OrderedDict
import shutil
import argparse
import json
import xmlschema
import yaml
from wfgenes.core.initial_wfgenes import BasewfGenes


class PywrapperwfGens(BasewfGenes):
    """Generate Python wrapper for wconfig nodes"""

    def __init__(self, wfgenes_init , blueargs):
        #BasewfGenes.__init__(wfgenes_init, blueargs)
        self.pywrapper_simstack(wfgenes_init)
    
    
    def pywrapper_simstack(self, wfgenes_init):

        first_i = 0
        for i in range(wfgenes_init.routine_number):
            wfg_task_wrapper = ""
            wfg_task_wrapper += "import yaml\n"
            wfg_task_wrapper += "import numpy as np\n"
            simstack_path = os.path.join(wfgenes_init.workflow_path, 'SimStack')
            wano_path = os.path.join(simstack_path, 'wanos', wfgenes_init.routine_name[i])
            if not os.path.exists(wano_path):
                os.makedirs(wano_path)
            wfgenes_init.wrapper_path = os.path.join(
                wano_path, wfgenes_init.routine_name[i] + '_wrapper.py')
            for j in range(wfgenes_init.subroutine_number[i]):
                if wfgenes_init.func[i][j][0] != 'BUILTIN' and wfgenes_init.func_nodup[i][j] != 'duplicate' and wfgenes_init.func_nodup[i][j] != 'duplicate':
                    wfg_task_wrapper += 'from ' + str(wfgenes_init.func[i][j][0]) + ' import ' + str(wfgenes_init.func[i][j][1]) + '\n'
                    first_i +=1
            wfg_task_wrapper += "\nif __name__ == '__main__':\n\n"
            for j in range(wfgenes_init.subroutine_number[i]):
                for k in range(len(wfgenes_init.inputs_nodup[i][j])):
                    if wfgenes_init.inputs_nodup[i][j][k] != 'duplicate' and wfgenes_init.inputs_no_locdep[i][
                            j][k] != 'inner_dependent' and wfgenes_init.inputs_links[i][j][k][0].isnumeric() == False:
                        wfg_task_wrapper += '\n \t#Read Input #' + str(k + 1) + ' from subroutine #' + str(j + 1) + ' in routine #' + str(i) + '\n'
                        wfg_task_wrapper += "\tyaml_stream = open('" + wfgenes_init.inputs[i][j][k] + ".yaml', 'r')\n"
                        wfg_task_wrapper += "\t" + wfgenes_init.inputs[i][j][k] + " = yaml.load(yaml_stream, Loader=yaml.Loader)\n\n"
                    if wfgenes_init.inputs_nodup[i][j][k] != 'duplicate' and wfgenes_init.inputs_no_locdep[i][j][k] != 'inner_dependent' and wfgenes_init.inputs_links[i][j][k][0].isnumeric():
                        wfg_task_wrapper += '\n \t#Read Input #' + str(k +1) + ' from subroutine #' + str(j +1) + ' in routine #' + str(i) + '\n'
                        wfg_task_wrapper += "\tyaml_stream = open('" + wfgenes_init.inputs_locname[i][j][k] + ".yaml', 'r')\n"
                        wfg_task_wrapper +=  "\t" + wfgenes_init.inputs_locname[i][j][k] + " = yaml.load(yaml_stream, Loader=yaml.Loader)\n\n"

                if wfgenes_init.func[i][j][0] != 'BUILTIN':
                    wfg_task_wrapper += '\n\t# Call subroutine #' + str(j + 1) + '\n'
                    wfg_task_wrapper += '\t'
                    for k in range(len(wfgenes_init.outputs[i][j])):
                        if k != len(wfgenes_init.outputs[i][j]) - 1:
                            wfg_task_wrapper += wfgenes_init.outputs_locname[i][j][k] + ", "
                        else:
                            wfg_task_wrapper += wfgenes_init.outputs_locname[i][j][k] + " = "
                    wfg_task_wrapper += wfgenes_init.func[i][j][1] + "("
                    for k in range(len(wfgenes_init.inputs[i][j])):
                        if k != len(
                                wfgenes_init.inputs[i][j]) - 1 and wfgenes_init.inputs_no_locdep[i][j][k] != 'inner_dependent' and wfgenes_init.inputs_links[i][j][k][0].isnumeric() == False:
                            wfg_task_wrapper += str(wfgenes_init.inputs[i][j][k]) + ', '
                        elif k != len(wfgenes_init.inputs[i][j]) - 1 and (wfgenes_init.inputs_no_locdep[i][j][k] == 'inner_dependent' or wfgenes_init.inputs_links[i][j][k][0].isnumeric()):
                            wfg_task_wrapper += str(wfgenes_init.inputs_locname[i][j][k]) + ', '
                        elif k == len(wfgenes_init.inputs[i][j]) - 1 and wfgenes_init.inputs_no_locdep[i][j][k] != 'inner_dependent' and wfgenes_init.inputs_links[i][j][k][0].isnumeric() == False:
                            wfg_task_wrapper += str(wfgenes_init.inputs[i][j][k]) + ")\n\n"
                        elif k == len(wfgenes_init.inputs[i][j]) - 1 and (wfgenes_init.inputs_no_locdep[i][j][k] == 'inner_dependent' or wfgenes_init.inputs_links[i][j][k][0].isnumeric()):
                            wfg_task_wrapper += str(wfgenes_init.inputs_locname[i][j][k]) + ")\n\n"

                    wfg_task_wrapper += "\n\t# Dump outputs into yaml files from subroutine #" + str(j + 1) + '\n'

                    for k in range(len(wfgenes_init.outputs[i][j])):
                        wfg_task_wrapper += '\n\t#Dump Output #' + str(k + 1) + ' from subroutine #' + str(j + 1) + ' in routine # ' + str(i) +'\n'
                        wfg_task_wrapper += "\tfile=open('" + wfgenes_init.outputs_locname[i][j][k] + ".yaml', 'w')\n"
                        wfg_task_wrapper +=  "\tyaml.dump(" + wfgenes_init.outputs_locname[i][j][k] + ", file, default_flow_style=False)\n"
                        wfg_task_wrapper += "\tfile.close\n"
                    wfg_task_wrapper += "\t#End of subroutine #" + str(j + 1) + ' in routine # ' + str(i) + '\n'

                elif wfgenes_init.func[i][j][0] == 'BUILTIN' and wfgenes_init.func[i][j][1] == 'MERGE':
                    wfg_task_wrapper += "\t# Merge Multiple Dictionaries \n"
                    for k in range(len(wfgenes_init.inputs[i][j])):
                        if wfgenes_init.inputs_nodup[i][j][k] != 'duplicate' and wfgenes_init.inputs_no_locdep[i][
                                j][k] != 'inner_dependent' and wfgenes_init.inputs_links[i][j][k][0].isnumeric() == False:
                            wfg_task_wrapper += "if isinstance(" + wfgenes_init.inputs[i][j][k] + ", np.number):\n"
                            wfg_task_wrapper += '\t' + wfgenes_init.inputs_locname[i][j][k] + ' = dict(' + wfgenes_init.inputs_locname[i][j][k] + '= float(' + wfgenes_init.inputs[i][j][k] + ')) \n'
                        elif wfgenes_init.inputs_nodup[i][j][k] != 'duplicate' and (wfgenes_init.inputs_no_locdep[i][j][k] == 'inner_dependent' or wfgenes_init.inputs_links[i][j][k][0].isnumeric()):
                            wfg_task_wrapper += "if isinstance(" + wfgenes_init.inputs_locname[i][j][k] + ", np.number):\n"
                            wfg_task_wrapper += '\t' + wfgenes_init.inputs_locname[i][j][k] + ' = dict(' + wfgenes_init.inputs_locname[i][j][k] + '= float(' + wfgenes_init.inputs_locname[i][j][k] + ')) \n'
                        if wfgenes_init.inputs_nodup[i][j][k] != 'duplicate' and wfgenes_init.inputs_no_locdep[i][
                                j][k] != 'inner_dependent' and wfgenes_init.inputs_links[i][j][k][0].isnumeric() == False:
                            wfg_task_wrapper += "else:\n"
                            wfg_task_wrapper += '\t' + wfgenes_init.inputs_locname[i][j][k] + ' = dict(' + wfgenes_init.inputs_locname[i][j][k] + '= ' + wfgenes_init.inputs[i][j][k] + ') \n\n'
                        elif wfgenes_init.inputs_nodup[i][j][k] != 'duplicate' and (wfgenes_init.inputs_no_locdep[i][j][k] == 'inner_dependent' or wfgenes_init.inputs_links[i][j][k][0].isnumeric()):
                            wfg_task_wrapper += "else:\n"
                            wfg_task_wrapper += '\t' + wfgenes_init.inputs_locname[i][j][k] + ' = dict(' + wfgenes_init.inputs_locname[i][j][k] + '= ' + wfgenes_init.inputs_locname[i][j][k] + ') \n\n'

                    wfg_task_wrapper += '\t' + wfgenes_init.outputs_locname[i][j][0] + '={'
                    for k in range(len(wfgenes_init.inputs[i][j])):
                        if wfgenes_init.inputs_nodup[i][j][k] != 'duplicate':
                            if k != len(wfgenes_init.inputs[i][j]) - 1:
                                wfg_task_wrapper += '**' + wfgenes_init.inputs_locname[i][j][k] + ','
                            elif k == len(wfgenes_init.inputs[i][j]) - 1:
                                wfg_task_wrapper += '**' + wfgenes_init.inputs_locname[i][j][k] + '}'
                    wfg_task_wrapper += '\n\n\t#Dump Output from Merged Dictionaries #\n'
                    wfg_task_wrapper += "\tfile=open('" + wfgenes_init.outputs_locname[i][j][0] + ".yaml', 'w')\n"
                    wfg_task_wrapper += "\tyaml.dump(" + wfgenes_init.outputs_locname[i][j][0] + ", file, default_flow_style=False)\n"
                    wfg_task_wrapper += "\tfile.close()\n"
                    wfg_task_wrapper += "\t#End of subroutine #" + str(j + 1) + ' in routine # ' + str(i) + '\n'
            with open(wfgenes_init.wrapper_path, 'w') as file:
                file.write(wfg_task_wrapper) 


